#
# $Id: bgp.tcl,v 1.7 2007/11/26 14:59:45 he Exp $
#
# Check status of BGP peer sessions
#

proc initBGP {} {

    ::persist::add ::bgpPeerOperState
    ::persist::add ::bgpPeerAdminState
    ::persist::add ::bgpPeerUpTime
    ::persist::add ::localAS
    ::persist::add ::bgpPeers
    ::persist::add ::sawPeer
}

proc bgpAdminDown { sh ix vbl } {

    foreach vb $vbl {
	set oid [lindex $vb 0]
	set val [lindex $vb 2]
	set var [lrange [split [mib name $oid] "."] 0 0]
	set $var $val
    }
    set r [name $sh]

    set old [eventIxExists $ix bgp]
    
    set eid [eventId $ix bgp]

    if { $old } {
	if { [getEventAttr $eid "bgpAS"] == $bgpPeerAdminStatus } {
	    return ;			# already noted
	}
    }

    setEventAttr $eid "router" $r
    setEventAttr $eid "polladdr" [::config::conf $r address]
    setEventAttr $eid "priority" [::config::conf $r priority]

    setEventAttr $eid "bgpOS" "down"
    setEventAttr $eid "bgpAS" $bgpPeerAdminStatus
    setEventAttr $eid "remote-addr" $bgpPeerRemoteAddr
    setEventAttr $eid "remote-AS" $bgpPeerRemoteAs
    setEventAttr $eid "peer-uptime" 0
    setEventAttr $eid "lastevent" "peer is admin turned off"

    set s [format "%s peer %s AS %s is turned off (%s)" \
	       $r $bgpPeerRemoteAddr $bgpPeerRemoteAs $bgpPeerAdminStatus]
    log $s
    eventLog $eid $s

    eventCommit $eid
}

proc bgpOperDown { sh ix vbl } {

    foreach vb $vbl {
	set oid [lindex $vb 0]
	set val [lindex $vb 2]
	set var [lrange [split [mib name $oid] "."] 0 0]
	set $var $val
    }
    set r [name $sh]

    set old [eventIxExists $ix bgp]

    set eid [eventId $ix bgp]

    if { $old } {
	if { [getEventAttr $eid "bgpOS"] == "down" } {
	    return ;			# already noted
	}
    }

    setEventAttr $eid "router" $r
    setEventAttr $eid "polladdr" [::config::conf $r address]
    setEventAttr $eid "priority" [::config::conf $r priority]

    setEventAttr $eid "bgpOS" "down"
    setEventAttr $eid "bgpAS" $bgpPeerAdminStatus
    setEventAttr $eid "remote-addr" $bgpPeerRemoteAddr
    setEventAttr $eid "remote-AS" $bgpPeerRemoteAs
    setEventAttr $eid "peer-uptime" $bgpPeerFsmEstablishedTime
    setEventAttr $eid "lastevent" "peer is down"

    set s [format "%s peer %s AS %s is down (%s)" \
	       $r $bgpPeerRemoteAddr $bgpPeerRemoteAs $bgpPeerState]
    log $s
    eventLog $eid $s

    eventCommit $eid
}

proc bgpExternalReset { sh ix vbl } {

    foreach vb $vbl {
	set oid [lindex $vb 0]
	set val [lindex $vb 2]
	set var [lrange [split [mib name $oid] "."] 0 0]
	set $var $val
    }
    set r [name $sh]

    set eid [eventId $ix bgp]
    setEventAttr $eid "router" $r
    setEventAttr $eid "polladdr" [::config::conf $r address]
    setEventAttr $eid "priority" [::config::conf $r priority]

    setEventAttr $eid "bgpOS" $bgpPeerState
    setEventAttr $eid "bgpAS" $bgpPeerAdminStatus
    setEventAttr $eid "remote-addr" $bgpPeerRemoteAddr
    setEventAttr $eid "remote-AS" $bgpPeerRemoteAs
    setEventAttr $eid "peer-uptime" $bgpPeerFsmEstablishedTime
    setEventAttr $eid "lastevent" "peer was reset (now up)"

    set s [format "%s peer %s AS %s was reset (now up)" \
	       $r $bgpPeerRemoteAddr $bgpPeerRemoteAs]
    log $s
    eventLog $eid $s

    eventCommit $eid
}

proc recordBGPpeer { sh uptime vbl } {
    global bgpPeerOperState
    global bgpPeerAdminState
    global localAS bgpPeerUpTime
    global bgpPeers sawPeer
    global lastPoll

    set lastPoll($sh) [clock seconds]

    foreach vb $vbl {
	set oid [lindex $vb 0]
	set val [lindex $vb 2]
	set var [lrange [split [mib name $oid] "."] 0 0]
	set $var $val
    }
    set r [name $sh]
    foreach v "bgpPeerState bgpPeerAdminStatus \
	bgpPeerLocalAddr bgpPeerRemoteAddr bgpPeerRemoteAs \
	bgpPeerFsmEstablishedTime" {
	if { ! [info exists $v] } {
	    log [format "router %s misses %s" $r $v]
	    log [format "router %s vbl %s" $r $vbl]
	    return
	}
    }
    set ix [format "%s,%s" $r $bgpPeerRemoteAddr]

#    log [format "debug: BGP r %s loc %s rem %s AS %s as %s os %s" \
	     $r $bgpPeerLocalAddr $bgpPeerRemoteAddr \
	     $bgpPeerRemoteAs $bgpPeerAdminStatus $bgpPeerState]

    # Bug in JunOS -- info from IPv6 BGP sessions spill over
    if { $bgpPeerRemoteAddr == "0.0.0.0" } { return }
    # Bug in earlier Cisco IOS, info from elsewhere (IPv6?) spills over
    if { $bgpPeerRemoteAddr == "32.1.7.0" } { return }

    if { $bgpPeerState == "established" } {
	if { [info exists bgpPeerUpTime($ix)] } {
	    # Ignore session resets as a result of a restart
	    if { $uptime >= $bgpPeerUpTime($ix) } {
		if { $bgpPeerUpTime($ix) > $bgpPeerFsmEstablishedTime } {
		    # peer was reset, but is now up
		    if { $localAS($sh) != $bgpPeerRemoteAs } {
			# external peer session was reset => alert
			bgpExternalReset $sh $ix $vbl
#			log [format "debug: Noted external reset for %s: %s" \
				 $r $ix]
		    }
		}
	    }
	}
    } else {
	if { $bgpPeerAdminStatus == "stop" } {
	    if { ! [info exists bgpPeerAdminState($ix)] } {
		set bgpPeerAdminState($ix) "unknown"
	    }
	    if { $bgpPeerAdminState($ix) != $bgpPeerAdminStatus } {
		# admin turned off
		bgpAdminDown $sh $ix $vbl
#		log [format "debug: Router %s peer %s AS %s admin-down" \
			 $r $bgpPeerRemoteAddr $bgpPeerRemoteAs]
	    } else {
#		log [format "debug: Router %s peer %s AS %s admin-down (but known)" \
			 $r $bgpPeerRemoteAddr $bgpPeerRemoteAs]
	    }
	} else {
	    if { ! [info exists bgpPeerOperState($ix)] } {
		set bgpPeerOperState($ix) "established"
	    }
	    # peer is not up, but configured to be up
	    # Check if it's an external AS
	    if { $localAS($sh) != $bgpPeerRemoteAs && \
		 $bgpPeerOperState($ix) == "established" } {
		# First verify that we've been up more than 10 minutes
		# before we flag it as an alert
		if { $uptime > 600 } {
		    bgpOperDown $sh $ix $vbl
#		    log [format "debug: Router %s peer %s AS %s os %s (down)" \
			     $r $bgpPeerRemoteAddr \
			     $bgpPeerRemoteAs $bgpPeerState]
		} else {
#		    log [format "debug: Router %s peer %s AS %s is %s (down), but uptime = %d" \
			     $r $bgpPeerRemoteAddr $bgpPeerRemoteAs \
			     $bgpPeerState $uptime]
		}
	    } else {
#		log [format "debug: Router %s peer %s AS %s os %s (down), but known not-up" \
			 $r $bgpPeerRemoteAddr \
			 $bgpPeerRemoteAs $bgpPeerState]
	    }
	}
    }
    set bgpPeerOperState($ix) $bgpPeerState
    set bgpPeerAdminState($ix) $bgpPeerAdminStatus
    set bgpPeerUpTime($ix) $bgpPeerFsmEstablishedTime


    # Possibly add to list of peers, record it as recently seen
    # to allow later garbage collection
    set found 0
    set now [clock seconds]
    if { [info exists bgpPeers($r)] } {
	foreach p $bgpPeers($r) {
	    if { $p == $bgpPeerRemoteAddr } {
		set found 1
		set k [format "%s,%s" $r $p]
		set sawPeer($k) $now
	    }
	}
    }
    if { ! $found } {
	lappend bgpPeers($r) $bgpPeerRemoteAddr
	set k [format "%s,%s" $r $bgpPeerRemoteAddr]
	set sawPeer($k) $now
    }

}

proc checkBGPpeers { sh up } {
    
    set vl "bgpPeerState bgpPeerAdminStatus \
	bgpPeerLocalAddr bgpPeerRemoteAddr bgpPeerRemoteAs \
	bgpPeerFsmEstablishedTime"
    sparseWalk $sh $vl x [list eval recordBGPpeer $sh $up {$x}]
}

proc checkBGP { sh } {
    global localAS

    $sh get "bgpLocalAs.0 sysUpTime.0" {
#	puts [format "%s %s %s" %S %E "%V"]
	if { "%E" != "noError" } {
	    return
	}
	set lasv [lindex "%V" 0]
	set supv [lindex "%V" 1]

	if { [lindex $lasv 1] == "noSuchObject" } {
	    return
	}
	set as [lindex $lasv 2]
	if { $as == 0 } {
	    return
	}
#	puts [format "localAS(%s) %s" %S $as]
	set localAS(%S) $as

	set up100 [mib scan [lindex $supv 0] [lindex $supv 2]]
	set up [expr $up100 / 100]

#	puts [format "%s uptime %d secs (%s)" %S $up [lindex $supv 2]]

	checkBGPpeers %S $up
    }
}
